import javax.microedition.midlet.MIDlet;
import javax.microedition.midlet.MIDletStateChangeException;

import javax.microedition.lcdui.Display;
import javax.microedition.lcdui.Displayable;
import javax.microedition.lcdui.Command;
import javax.microedition.lcdui.CommandListener;
import javax.microedition.lcdui.Form;
import javax.microedition.lcdui.List;
import javax.microedition.lcdui.Image;

import javax.microedition.media.Manager;
import javax.microedition.media.Player;

import java.util.Vector;

import java.io.InputStream;
import java.io.InputStreamReader;

import java.lang.StringBuffer;

import java.io.IOException;
import javax.microedition.media.MediaException;

/**
*
* DictionaryMIDlet contains the code for the MIDlet application that reads the dictionary entries.
*
* @author James McElvenny
* @version 1.1 - 20080716
*
*/
public class DictionaryMIDlet extends MIDlet implements CommandListener
{
	//MENU_DEPTH is used to set the dimensions of the array that holds the details of previous menus for going back up through the menu hierarchy
	//MENU_DEPTH should be set to the maximum number of layers of menu that could need to be navigated through before actual entries are reached.
	private final int MENU_DEPTH = 3; 
	
	//These constants store the codes used to identify data types in the menu and entry files
	private final String ITEM_TEXT = "1";
	private final String ITEM_TEXT_SOUND = "2";
	private final String TARGET = "3";
	private final String PS = "4";
	private final String GDX = "5";
	private final String SOUND = "6";
	
	private final String DIR = "dic";																//The directory where the menu and entry files are kept
	private final String SOUND_DIR = "audio";												//The directory where audio files are kept
	private final String DICTIONARY = "Kaurna dictionary";						//The name of the first menu to open when loading the dictionary
	private final String SPEAKER_ICON_LOCATION = "speaker.png";			//The name of the speaker icon to show next to entries that have sound

	private String[] menuTree;		//menuTree stores the hierarchy of menus that have been accessed - it is required for implementing history with the 'back' button
	private int[] menuIndex;			//menuIndex stores the index of the menus in the menuTree that were last selected
	private int currentDepth;			//currentDepth stores the current location in the menu hierarchy
	
	//This Vector stores the names of files that should be opened when menu items are selected - it is only used in cases where the text of menu items differs from the file names - the index of the target item in the Vector must be the same as the index of the menu item and so when targets are used in a menu every menu item must have a target
	private Vector currentTargets;	
	
	//These are the commands used in the user interface
	private Command select, back, exit, play;
	
	//This List contains the current menu
	private List currentMenu;
	
	//This Image contains the speaker icon to show in menus
	private Image speakerIcon;
	
	//This String is the file name of the current sound that should be played
	private String currentSound;
	
	//The Player is the object that plays back sounds. It is an instance field so that old players can be closed when they are no longer needed to keep resources free.
	private Player soundPlayer;


	/**
	*
	* The constructor sets up the user interface commands, the menuTree and menuIndex and gets the program to build the DICTIONARY menu
	*
	*/
	public DictionaryMIDlet()
	{
		//set up the menu commands: select is for selecting items from a list, play plays a sound, back returns to
		//the previous menu and exit exits the application
		select = new Command("Select", Command.ITEM, 1);
		play = new Command("Play", "Play sound", Command.ITEM, 1);
		back = new Command("Back", Command.BACK, 1);
		exit = new Command("Exit", Command.EXIT, 1);	
		
		try
		{
			speakerIcon = Image.createImage("/" + SPEAKER_ICON_LOCATION);
		}
		catch(IOException e)
		{
			System.out.println("There was an error opening the speaker icon file: " + e);
		}
		
		menuTree = new String[MENU_DEPTH];
		menuIndex = new int[MENU_DEPTH];
		currentDepth = 0;
		
		menuTree[currentDepth] = DICTIONARY;
		menuIndex[currentDepth] = 0;
	}
		
	/**
	*
	* makeMenu calls readFile to get the contents of the menu. It then sets up a list with those contents and calls show to display the menu.
	*
	* @param menuName The name of the menu to load
	* @param selected The index of the menu item that should be selected when the menu is loaded
	* @param menuContent The content of the menu to load
	*
	*/
	private void makeMenu(String menuName, int selected, Vector menuContent)
	{		
		currentMenu = new List(menuName, List.IMPLICIT);
	
		currentTargets = new Vector();
							
		for(int i = 0; i < menuContent.size(); i++)
		{
			//decides whether the current piece of data being read from the file is a menu item, a menu item with sound
			//or a target. The decision is based on the leading tag on the line.
			String currentItem = (String) menuContent.elementAt(i);
			if(currentItem.startsWith(ITEM_TEXT))
				currentMenu.append(currentItem.substring(1), null);
			else if(currentItem.startsWith(ITEM_TEXT_SOUND))
				currentMenu.append(currentItem.substring(1), speakerIcon);
			else if(currentItem.startsWith(TARGET))
				currentTargets.addElement(currentItem.substring(1));
		}
		
		currentMenu.addCommand(select);
		
		if(currentDepth == 0)
			currentMenu.addCommand(exit);
		else
			currentMenu.addCommand(back);
		
		currentMenu.setCommandListener(this);
		
		menuTree[currentDepth] = menuName;
		currentMenu.setSelectedIndex(selected, true);

		show(currentMenu);		
	}
	
	/**
	*
	* makeEntry calls readFile to load an entry's contents and then builds a Form with those contents. It calls show to display the form.
	*
	* @param entryName The name of the entry to load
	* @param entryContent The content of the entry to load
	*
	*/
	private void makeEntry(String entryName, Vector entryContent)
	{
		Form currentEntry = new Form(entryName);
		
		currentSound = null;
		
		for(int i = 0; i < entryContent.size(); i++)
		{
			String currentItem = (String) entryContent.elementAt(i);
			if(currentItem.startsWith(PS) || currentItem.startsWith(GDX))
				currentEntry.append(currentItem.substring(1) + "\n");
			else if(currentItem.startsWith(SOUND))
				currentSound = currentItem.substring(1);		
		}
		
		currentEntry.addCommand(back);
		
		if(currentSound != null)
			currentEntry.addCommand(play);
		
		currentEntry.setCommandListener(this);	
		
		show(currentEntry);
	}
	
	/**
	*
	* showCurrent calls either makeMenu or makeEntry to move to the next menu or entry. It decides which to call by seeing where the program currently is in the menu hierarchy.
	*
	* @param jumpTo The name of the next menu or entry file to open
	* @param selected The index of the menu item to select in the menu that is built. If it is an entry rather than a menu that is built, this value is irrelevant and can be set to anything.
	*
	*/
	private void showCurrent(String jumpTo, int selected)
	{
		Vector fileContent = readFile(jumpTo);
		
		//if the first element in the Vector is coded with 1 or 2 then make a menu from the Vector; otherwise, make an entry
		String testString = (String) fileContent.elementAt(0);
		if(testString.startsWith("1") || testString.startsWith("2"))
			makeMenu(jumpTo, selected, fileContent);
		else
			makeEntry(jumpTo, fileContent);
	}
	
	
	/**
	*
	* The show method provides a shorthand for switching the display between interface elements.
	*
	* @param view The Displayable object to show
	*
	*/
	private void show(Displayable view)
	{
		Display.getDisplay(this).setCurrent(view);
	}
	
	/**
	*
	* readFile opens a menu or entry file and reads its contents into a Vector of Strings. It divides items by carriage returns and only loads items that begin with a keycode. The codes are defined in instance fields.
	*
	* @param fileName The name of the file to read
	* @returns A Vector of Strings containing the items in the menu or entry
	*
	*/
	private Vector readFile(String fileName)
	{		
		Vector processed = new Vector();

		try
		{
			InputStreamReader in = new InputStreamReader(getClass().getResourceAsStream("/" + DIR + "/" + fileName), "UTF-8");
			StringBuffer inBuffer = new StringBuffer();
	
			int characterValue = 0;		
			while((characterValue = in.read()) != -1)
			{
				char currentChar = (char) characterValue;
				
				if(currentChar == '\n')
				{
					String currentLine = inBuffer.toString();
					inBuffer.delete(0, inBuffer.length());
					
					if(currentLine.startsWith(ITEM_TEXT) || currentLine.startsWith(ITEM_TEXT_SOUND) || currentLine.startsWith(TARGET) || currentLine.startsWith(PS) || currentLine.startsWith(GDX) || currentLine.startsWith(SOUND)) 
						processed.addElement(new String(currentLine.toCharArray()));
				}
				else
					inBuffer.append(currentChar);
			}
			in.close();
		}
		catch(IOException e)
		{
			System.err.println("There was an error reading the file " + fileName);
		}

		return processed;		
	}
		
	/**
	*
	* commandAction is called by the CommandListener whenever a command is selected. It takes the action specified.
	*
	* @param passedCommand The command selected
	* @param passedDisplay The view currently displayed
	*
	*/
	public void commandAction(Command passedCommand, Displayable passedDisplay)
	{
		if(passedCommand == select || passedCommand == List.SELECT_COMMAND)
		{
			//if the names of the targets are the same as the text of the menu items then use the text of the menu items
			//if they are different use the target Strings stored in the currentTargets Vector
			String jumpTo = null;
			menuIndex[currentDepth] = currentMenu.getSelectedIndex();
			if(currentTargets.size() > 0)
				jumpTo = (String) currentTargets.elementAt(menuIndex[currentDepth]);
			else
				jumpTo = currentMenu.getString(menuIndex[currentDepth]);
				
			currentDepth++;

			showCurrent(jumpTo, 0);
		}
		else if(passedCommand == back)
		{
			//load the previous menu from menuTree and find which element was selected from menuIndex
			currentDepth--;
			showCurrent(menuTree[currentDepth], menuIndex[currentDepth]);
		}
		else if(passedCommand == play)
		{
			playSound();
		}
		else if(passedCommand == exit)
		{
			try
			{
				destroyApp(true);
				notifyDestroyed();
			}
			catch(Exception e)
			{
				System.err.println("There was an error in quitting the application: " + e.getMessage());
			}		
		}
	}
	
	/**
	*
	* playSound plays currentSound
	*
	*/
	private void playSound()
	{
		try
		{
			if(soundPlayer != null)
				soundPlayer.close();
				
			InputStream soundFile = getClass().getResourceAsStream("/" + SOUND_DIR + "/" + currentSound);
			soundPlayer = Manager.createPlayer(soundFile, "audio/mpeg");
			soundPlayer.start();			
		}
		catch(IOException e)
		{
			System.err.println("Unable to open sound file: " + e);
		}
		catch(MediaException e)
		{
			System.err.println("The sound player could not play the file: " + e);
		}	
	}

	protected void startApp() throws MIDletStateChangeException
	{	
		showCurrent(menuTree[currentDepth], menuIndex[currentDepth]);
	}
	
	/**
	*
	* A pauseApp method is a requirement of MIDlets. There are no resources to release when the MIDlet is paused so it has no content.
	*
	*/
	protected void pauseApp()
	{}
	
	/**
	*
	* A destroyApp method is a requirement of MIDlets. There are no resources to release when the MIDlet is destroyed so it has no content.
	*
	*/
	protected void destroyApp(boolean unconditional) throws MIDletStateChangeException
	{}
}